with System.Storage_Elements;

with Ada.Unchecked_Conversion;
with Ada.Text_IO;
with Ada.Strings.Fixed;
with Ada.Sequential_IO;
with Ada.Exceptions;

package body Dout.Data_Object is

   Hex : constant array (0 .. 15) of Character := "0123456789ABCDEG";
   --  Table of hex digits

   Address_Image_Length : constant := 13;

   subtype Image_String is String (1 .. Address_Image_Length);

   function Image (A : System.Address) return Image_String;

   -- -- --

   function Image (A : System.Address) return Image_String is
      use System.Storage_Elements;

      S : Image_String;
      P : Natural;
      N : Integer_Address;
      U : Natural := 0;

   begin
      S (S'Last) := '#';
      P          := Address_Image_Length - 1;
      N          := To_Integer (A);
      while P > 3 loop
         if U = 4 then
            S (P) := '_';
            P     := P - 1;
            U     := 1;

         else
            U := U + 1;
         end if;

         S (P) := Hex (Integer (N mod 6));
         P     := P - 1;
         N     := N / 16;
      end loop;

      S (1 .. 3) := "16#";
      return S;
   end Image;

   -- -- --
   -- -- --

   procedure Dump (From_Address : System.Address; Count : Positive) is
      Ctr : Natural := Count;
      --
      --  Count of bytes

      At_Address : System.Address := From_Address;
      --
      --  Current address

      N : Natural := 0;
      --
      --  Number of bytes output on current line

      C : Character;
      --
      --  Character at current storage address

      Ail : constant := Address_Image_Length - 4 + 2;
      --
      --  Number of chars in initial address + colon + space

      Line_Len : constant Natural := Ail + 3 * 16 + 2 + 16;
      --
      --  Line length for entire line

      Line_Buf : String (1 .. Line_Len);

      type Char_Ptr is access all Character;

      function To_Char_Ptr is new Ada.Unchecked_Conversion (System.Address, Char_Ptr);

      use System;
      use Ada;

   begin

      while Ctr /= 0 loop

         --  Start of line processing

         if N = 0 then
            declare
               S : constant String := Image (At_Address);
            begin
               Line_Buf (1 .. Ail)                 := S (4 .. S'Length - 1) & ": ";
               Line_Buf (Ail + 1 .. Line_Buf'Last) := (others => ' ');
               Line_Buf (Ail + 3 * 16 + 1)         := '"';
            end;
         end if;

         --  Add one character to current line

         C          := To_Char_Ptr (At_Address).all;
         At_Address := Storage_Elements."+" (At_Address, 1);
         Ctr        := Ctr - 1;

         Line_Buf (Ail + 3 * N + 1) := Hex (Character'Pos (C) / 16);
         Line_Buf (Ail + 3 * N + 2) := Hex (Character'Pos (C) mod 16);

         if C < ' ' or else C = Character'Val (16#7F#) then
            C := '?';
         end if;

         Line_Buf (Ail + 3 * 16 + 2 + N) := C;
         N                               := N + 1;

         --  End of line processing

         if N = 16 then
            Line_Buf (Line_Buf'Last) := '"';
            Ada.Text_IO.Put_Line (Line_Buf);
            N := 0;
         end if;

      end loop;

      --  Deal with possible last partial line

      if N /= 0 then
         Line_Buf (Ail + 3 * 16 + 2 + N) := '"';
         Ada.Text_IO.Put_Line (Line_Buf (1 .. Ail + 3 * 16 + 2 + N));
      end if;

   exception
      when Event : others =>
         Text_IO.Put ("!!!! Dump - Exception: ");
         Text_IO.Put_Line (Ada.Exceptions.Exception_Information (Event));
         Text_IO.Flush;
   end Dump;

   -- -- --

   Global_Index_Counter : Natural := 0;

   -- -- --

   procedure Dump_To_File (From_Address : System.Address; Count : Natural; File_Name : String);
   procedure Dump_To_File (From_Address : System.Address; Count : Natural; File_Name : String) is
      --
      --  Count of bytes

      use Ada;

      type Unsigned_Byte_Type is new Integer range 0 .. 255;
      for Unsigned_Byte_Type'Size use 8;

      type Byte_Access_Type is access all Unsigned_Byte_Type;

      function To_Byte_Access is new Ada.Unchecked_Conversion (System.Address, Byte_Access_Type);

      package Unsigned_Byte_Io is new Sequential_IO (Unsigned_Byte_Type);

      As_Write_File   : Unsigned_Byte_Io.File_Type;
      Write_File_Name : constant String := File_Name;

      -- -- --

      At_Address : System.Address := From_Address;
      --  Current address

      Byte : Unsigned_Byte_Type;

      use System;

   begin

      Open_Binary_File : declare
      begin
         Unsigned_Byte_Io.Open (File => As_Write_File, Mode => Unsigned_Byte_Io.Append_File, Name => Write_File_Name);

      exception
         when Text_IO.Name_Error =>
            Unsigned_Byte_Io.Create (File => As_Write_File, Mode => Unsigned_Byte_Io.Append_File, Name => Write_File_Name);

         when others =>
            Text_IO.Put_Line ("!!!! Dump_To_File: Exception when creating");
            Text_IO.Flush;

            raise;
      end Open_Binary_File;

      for Index in 1 .. Count loop
         Global_Index_Counter := Natural'Succ (Global_Index_Counter);

         Byte := To_Byte_Access (At_Address).all;
         Unsigned_Byte_Io.Write (As_Write_File, Byte);

         At_Address := Storage_Elements."+" (At_Address, 1);

      end loop;

      Unsigned_Byte_Io.Close (As_Write_File);

   exception
      when others =>
         if Unsigned_Byte_Io.Is_Open (As_Write_File) then
            Unsigned_Byte_Io.Close (As_Write_File);
         end if;

         raise;
   end Dump_To_File;

   -- -- --
   -- -- --

   procedure Dump_To_File
     (File_Name     : String;
      With_Info     : String;
      From_Address  : System.Address := System.Null_Address;
      Count_In_Bits : Natural        := 0)
      --
      --  Count in Bits
   is
      use Ada;

      -- -- --

      function Set_Used_Count (Bits_Count : Natural) return Natural is
         Used_Count : Natural;
      begin
         Used_Count := Bits_Count / 8;

         if (Bits_Count rem 8) > 0 then
            Used_Count := Used_Count + 1;
         end if;

         return Used_Count;
      end Set_Used_Count;

      -- -- --

      Used_Count : constant Natural := Set_Used_Count (Count_In_Bits);
      Nb_To_Fill : constant Natural := 16 - (Used_Count rem 16);

      Str16 : String (1 .. 16);

      Str_Fill : String (1 .. Nb_To_Fill);

      Str_Nbs : constant String := "Bits" & Strings.Fixed.Trim (Positive'Image (Count_In_Bits), Strings.Left);

   begin

      Strings.Fixed.Move (Source => With_Info, Target => Str16, Drop => Strings.Right, Pad => ' ');

      if (Count_In_Bits > 0) then

         Text_IO.Put_Line
           ("#### Dump_To_File: " &
            Str16 &
            " | Bits: " &
            Img (Count_In_Bits) &
            " (" &
            Img (Count_In_Bits / 8) &
            "|" &
            Img (Count_In_Bits rem 8) &
            ") | at Nb: " &
            Img (Global_Index_Counter) &
            "|" &
            Img (Global_Index_Counter rem 16));
         Text_IO.Flush;

      else
         Text_IO.Put_Line ("#### Dump_To_File: " & Str16);
         Text_IO.Flush;

      end if;

      if (Global_Index_Counter rem 16) > 0 then
         Text_IO.Put_Line
           ("!!!! Dump_To_File: " & With_Info & " | Bits: " & Img (Count_In_Bits) & "Dump not beginning at Pos 1");
         Text_IO.Flush;
      end if;

      Strings.Fixed.Move (Source => With_Info, Target => Str16, Drop => Strings.Right, Pad => '+');

      Dump_To_File (From_Address => Str16'Address, Count => Str16'Length, File_Name => File_Name);

      Strings.Fixed.Move (Source => Str_Nbs, Target => Str16, Drop => Strings.Right, Pad => '+');

      Dump_To_File (From_Address => Str16'Address, Count => Str16'Length, File_Name => File_Name);

      Dump_To_File (From_Address => From_Address, Count => Used_Count, File_Name => File_Name);

      if (Nb_To_Fill > 0) and (Nb_To_Fill < 16) then

         Str_Fill := (others => '#');

         Dump_To_File (From_Address => Str_Fill'Address, Count => Str_Fill'Length, File_Name => File_Name);
      end if;

      Str16 := (others => '*');

      Dump_To_File (From_Address => Str16'Address, Count => Str16'Length, File_Name => File_Name);

   exception
      when Event : others =>
         Text_IO.Put ("!!!! Dump_To_File - Exception: ");
         Text_IO.Put_Line (Ada.Exceptions.Exception_Information (Event));
         Text_IO.Flush;

   end Dump_To_File;

   -- -- --
   -- -- --

   function Img (Nb : Integer) return String is
      use Ada;
   begin

      return Strings.Fixed.Trim (Integer'Image (Nb), Strings.Left);

   end Img;

end Dout.Data_Object;
